/*
* dvb-wss-overdrive.c:  Tool to overdrive dvb ff (av7110) WSS signaling
*
* Copyright (C) 2006 Jens Vogel (jens.vogel@justmail.de)
*
* Based on wsstest.c:  Tool to test av7110 WSS interface - Rev. 0.1
* from Oliver Endriss (o.endriss@gmx.de)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
* Or, point your browser to http://www.gnu.org/copyleft/gpl.html
*/

#include <stdio.h>
#include <unistd.h>
#include <memory.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/videodev2.h>
#include <errno.h>
#include <stdlib.h>

#define DEFAULT_DEVICE "/dev/vbi0"

typedef struct 
{
  unsigned    iWssData;
  const char* pszName;
} SWssInfo;
SWssInfo aWssInfo[] =
{
  { 0x7, "16:9" },
  { 0x8, "4:3" },
  { 0x1, "L14:9" },
  { 0xb, "L16:9" },
  { 0xd, "L>16:9" }
};
const int naWssInfo = sizeof(aWssInfo) / sizeof(SWssInfo);

int OpenDevice(const char* szDevice)
{
  int device = open(szDevice, O_WRONLY);
  if (device < 0) 
  {
    fprintf(stderr, "Error: Can't open device '%s' (%s)\n", szDevice, strerror(errno));
    exit(1);
  }

  struct v4l2_format fmt;
	memset(&fmt, 0, sizeof fmt);
	fmt.type = V4L2_BUF_TYPE_SLICED_VBI_OUTPUT;
	fmt.fmt.sliced.service_set = V4L2_SLICED_WSS_625;
	if (ioctl(device, VIDIOC_S_FMT, &fmt) < 0) {
    fprintf(stderr, "Error: Can't init device '%s' (%s)\n", szDevice, strerror(errno));
		close(device);
		exit(2);
	}

  return device;
}

int main(int argc, char *argv[])
{
  if (argc > 2) 
  {
    fprintf(stderr, 
            "Usage: %s [<dev>]\n"
            "  Default for <dev> is " DEFAULT_DEVICE, 
            argv[0]);
    return 99;
  }

  const char* szDevice = (argc > 1) ? argv[1] : DEFAULT_DEVICE;
  
  close(OpenDevice(szDevice)); // test the device

  char szInput[100];
  int device = -1;
  while (!ferror(stdin))
    if (scanf("%s", szInput) != 1)
      usleep(100000);
    else
    {
      if (!strcmp(szInput, "exit"))
      {
        printf("> exit\n");
        if (device >= 0)
          close(device);
        return 0;
      }
      else if (!strcmp(szInput, "free"))
      {
        if (device >= 0)
        {
          printf("> free device\n");
          close(device);
          device = -1;
        }
      }
      else
      {
        int      bWssData = 0;
        unsigned iWssData;
        for (int i = 0; (i < naWssInfo) && !bWssData; i++)
          if (!strcmp(aWssInfo[i].pszName, szInput))
          {
            bWssData = 1;
            iWssData = aWssInfo[i].iWssData;
          }
        if (!bWssData)
          bWssData = sscanf(szInput, "%x", &iWssData) == 1;

        if (!bWssData)
          fprintf(stderr, "Error: Unkown input '%s'; ignored\n", szInput);
        else
        {
          if (device < 0)
            device = OpenDevice(szDevice);

          bWssData = 0;
          for (int i = 0; (i < naWssInfo) && !bWssData; i++)
            if (iWssData == aWssInfo[i].iWssData)
            {
              bWssData = 1;
              printf("> WSS 0x%x = %s\n", iWssData, aWssInfo[i].pszName);
            }
          if (!bWssData)
            printf("> WSS 0x%x\n", iWssData);
          fflush(stdout);
        
          struct v4l2_sliced_vbi_data data;
          memset(&data, 0, sizeof data);
          data.id = (iWssData) ? V4L2_SLICED_WSS_625 : 0;
          data.line = 23;
          data.data[0] = iWssData & 0xff;
          data.data[1] = (iWssData >> 8) & 0x3f;
          if (write(device, &data, sizeof data) != sizeof data) 
          {
            fprintf(stderr, "Error: Can't write to device '%s' (%s)\n", szDevice, strerror(errno));
            if (device >= 0)
              close(device);
            return 4;
          }
        }
      }
    }

  fprintf(stderr, "Error: Can't read input (%s)\n", strerror(errno));
  if (device >= 0)
    close(device);
  return 1;
}
